# Copyright (C) 2018 by Kevin L. Mitchell <klmitch@mit.edu>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you
# may not use this file except in compliance with the License. You may
# obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied. See the License for the specific language governing
# permissions and limitations under the License.

import functools
import inspect

import pytest
import six

from micropath import injector


class TestWantSignature(object):
    if six.PY2:
        def test_py2_getsig_noargs(self, mocker):
            argspec = inspect.ArgSpec([], None, None, None)
            mock_getargspec = mocker.patch.object(
                injector.inspect, 'getargspec',
                return_value=argspec,
            )

            result = injector.WantSignature._getsig('func')

            assert result == (
                [],
                set(),
                set(),
                False,
                False,
            )
            mock_getargspec.assert_called_once_with('func')

        def test_py2_getsig_withargs_nodefaults(self, mocker):
            argspec = inspect.ArgSpec(['a', 'b', 'c'], None, None, None)
            mock_getargspec = mocker.patch.object(
                injector.inspect, 'getargspec',
                return_value=argspec,
            )

            result = injector.WantSignature._getsig('func')

            assert result == (
                ['a', 'b', 'c'],
                set(['a', 'b', 'c']),
                set(),
                False,
                False,
            )
            mock_getargspec.assert_called_once_with('func')

        def test_py2_getsig_withargs_withdefaults(self, mocker):
            argspec = inspect.ArgSpec(['a', 'b', 'c'], None, None, (2, 3))
            mock_getargspec = mocker.patch.object(
                injector.inspect, 'getargspec',
                return_value=argspec,
            )

            result = injector.WantSignature._getsig('func')

            assert result == (
                ['a', 'b', 'c'],
                set(['a']),
                set(['b', 'c']),
                False,
                False,
            )
            mock_getargspec.assert_called_once_with('func')

        def test_py2_getsig_allposargs(self, mocker):
            argspec = inspect.ArgSpec(['a', 'b', 'c'], 'd', None, (2, 3))
            mock_getargspec = mocker.patch.object(
                injector.inspect, 'getargspec',
                return_value=argspec,
            )

            result = injector.WantSignature._getsig('func')

            assert result == (
                ['a', 'b', 'c'],
                set(['a']),
                set(['b', 'c']),
                True,
                False,
            )
            mock_getargspec.assert_called_once_with('func')

        def test_py2_getsig_allkwargs(self, mocker):
            argspec = inspect.ArgSpec(['a', 'b', 'c'], None, 'e', (2, 3))
            mock_getargspec = mocker.patch.object(
                injector.inspect, 'getargspec',
                return_value=argspec,
            )

            result = injector.WantSignature._getsig('func')

            assert result == (
                ['a', 'b', 'c'],
                set(['a']),
                set(['b', 'c']),
                False,
                True,
            )
            mock_getargspec.assert_called_once_with('func')
    else:  # Python 3
        def test_py3_getsig_noargs(self, mocker):
            signature = inspect.Signature(parameters=[])
            mock_signature = mocker.patch.object(
                injector.inspect, 'signature',
                return_value=signature,
            )

            result = injector.WantSignature._getsig('func')

            assert result == (
                [],
                set(),
                set(),
                False,
                False,
            )
            mock_signature.assert_called_once_with(
                'func',
                follow_wrapped=False,
            )

        def test_py3_getsig_withargs_nodefaults(self, mocker):
            signature = inspect.Signature(parameters=[
                inspect.Parameter('a', inspect.Parameter.POSITIONAL_ONLY),
                inspect.Parameter(
                    'b', inspect.Parameter.POSITIONAL_OR_KEYWORD,
                ),
                inspect.Parameter('c', inspect.Parameter.KEYWORD_ONLY),
            ])
            mock_signature = mocker.patch.object(
                injector.inspect, 'signature',
                return_value=signature,
            )

            result = injector.WantSignature._getsig('func')

            assert result == (
                ['a', 'b'],
                set(['b', 'c']),
                set(),
                False,
                False,
            )
            mock_signature.assert_called_once_with(
                'func',
                follow_wrapped=False,
            )

        def test_py3_getsig_withargs_withdefaults(self, mocker):
            signature = inspect.Signature(parameters=[
                inspect.Parameter('a', inspect.Parameter.POSITIONAL_ONLY),
                inspect.Parameter(
                    'b', inspect.Parameter.POSITIONAL_OR_KEYWORD,
                ),
                inspect.Parameter(
                    'c', inspect.Parameter.POSITIONAL_OR_KEYWORD,
                    default=3,
                ),
                inspect.Parameter(
                    'd', inspect.Parameter.KEYWORD_ONLY,
                    default=4,
                ),
            ])
            mock_signature = mocker.patch.object(
                injector.inspect, 'signature',
                return_value=signature,
            )

            result = injector.WantSignature._getsig('func')

            assert result == (
                ['a', 'b', 'c'],
                set(['b']),
                set(['c', 'd']),
                False,
                False,
            )
            mock_signature.assert_called_once_with(
                'func',
                follow_wrapped=False,
            )

        def test_py3_getsig_allposargs(self, mocker):
            signature = inspect.Signature(parameters=[
                inspect.Parameter('a', inspect.Parameter.POSITIONAL_ONLY),
                inspect.Parameter(
                    'b', inspect.Parameter.POSITIONAL_OR_KEYWORD,
                ),
                inspect.Parameter(
                    'c', inspect.Parameter.POSITIONAL_OR_KEYWORD,
                    default=3,
                ),
                inspect.Parameter('d', inspect.Parameter.VAR_POSITIONAL),
                inspect.Parameter(
                    'e', inspect.Parameter.KEYWORD_ONLY,
                    default=4,
                ),
            ])
            mock_signature = mocker.patch.object(
                injector.inspect, 'signature',
                return_value=signature,
            )

            result = injector.WantSignature._getsig('func')

            assert result == (
                ['a', 'b', 'c'],
                set(['b']),
                set(['c', 'e']),
                True,
                False,
            )
            mock_signature.assert_called_once_with(
                'func',
                follow_wrapped=False,
            )

        def test_py3_getsig_allkwargs(self, mocker):
            signature = inspect.Signature(parameters=[
                inspect.Parameter('a', inspect.Parameter.POSITIONAL_ONLY),
                inspect.Parameter(
                    'b', inspect.Parameter.POSITIONAL_OR_KEYWORD,
                ),
                inspect.Parameter(
                    'c', inspect.Parameter.POSITIONAL_OR_KEYWORD,
                    default=3,
                ),
                inspect.Parameter(
                    'd', inspect.Parameter.KEYWORD_ONLY,
                    default=4,
                ),
                inspect.Parameter('e', inspect.Parameter.VAR_KEYWORD),
            ])
            mock_signature = mocker.patch.object(
                injector.inspect, 'signature',
                return_value=signature,
            )

            result = injector.WantSignature._getsig('func')

            assert result == (
                ['a', 'b', 'c'],
                set(['b']),
                set(['c', 'd']),
                False,
                True,
            )
            mock_signature.assert_called_once_with(
                'func',
                follow_wrapped=False,
            )

    def test_from_func_cached(self, mocker):
        mock_getsig = mocker.patch.object(
            injector.WantSignature, '_getsig',
            return_value=(
                ['a', 'b', 'c'], set(['a']), set(['b', 'c']),
                False, False,
            ),
        )
        mock_init = mocker.patch.object(
            injector.WantSignature, '__init__',
            return_value=None,
        )
        func = mocker.Mock(_micropath_signature='signature')

        result = injector.WantSignature.from_func(func)

        assert result == 'signature'
        assert func._micropath_signature == 'signature'
        mock_getsig.assert_not_called()
        mock_init.assert_not_called()

    def test_from_func_uncached_base(self, mocker):
        mock_getsig = mocker.patch.object(
            injector.WantSignature, '_getsig',
            return_value=(
                ['a', 'b', 'c'], set(['a']), set(['b', 'c']),
                False, False,
            ),
        )
        mock_init = mocker.patch.object(
            injector.WantSignature, '__init__',
            return_value=None,
        )
        func = mocker.Mock(spec=[])

        result = injector.WantSignature.from_func(func)

        assert isinstance(result, injector.WantSignature)
        assert func._micropath_signature is result
        mock_getsig.assert_called_once_with(func)
        mock_init.assert_called_once_with(
            func, ['a', 'b', 'c'], set(['a']), set(['b', 'c']),
            False, False,
        )

    def test_from_func_uncached_not_allkw(self, mocker):
        mock_getsig = mocker.patch.object(
            injector.WantSignature, '_getsig',
            return_value=(
                ['a', 'b', 'c'], set(['a']), set(['b', 'c']),
                False, False,
            ),
        )
        mock_init = mocker.patch.object(
            injector.WantSignature, '__init__',
            return_value=None,
        )
        func = mocker.Mock(spec=[])

        result = injector.WantSignature.from_func(
            func,
            provides=['d'],
            required=['e'],
            optional=['f'],
        )

        assert isinstance(result, injector.WantSignature)
        assert func._micropath_signature is result
        mock_getsig.assert_called_once_with(func)
        mock_init.assert_called_once_with(
            func, ['a', 'b', 'c'], set(['a']), set(['b', 'c']),
            False, False,
        )

    def test_from_func_uncached_with_allkw_and_req_opt(self, mocker):
        mock_getsig = mocker.patch.object(
            injector.WantSignature, '_getsig',
            return_value=(
                ['a', 'b', 'c'], set(['a']), set(['b', 'c']),
                False, True,
            ),
        )
        mock_init = mocker.patch.object(
            injector.WantSignature, '__init__',
            return_value=None,
        )
        func = mocker.Mock(spec=[])

        result = injector.WantSignature.from_func(
            func,
            provides=['d'],
            required=['e'],
            optional=['f'],
        )

        assert isinstance(result, injector.WantSignature)
        assert func._micropath_signature is result
        mock_getsig.assert_called_once_with(func)
        mock_init.assert_called_once_with(
            func, ['a', 'b', 'c'], set(['a', 'e']), set(['b', 'c', 'f']),
            False, False,
        )

    def test_from_func_uncached_with_allkw_without_req_opt(self, mocker):
        mock_getsig = mocker.patch.object(
            injector.WantSignature, '_getsig',
            return_value=(
                ['a', 'b', 'c'], set(['a']), set(['b', 'c']),
                False, True,
            ),
        )
        mock_init = mocker.patch.object(
            injector.WantSignature, '__init__',
            return_value=None,
        )
        func = mocker.Mock(spec=[])

        result = injector.WantSignature.from_func(
            func,
            provides=['d'],
        )

        assert isinstance(result, injector.WantSignature)
        assert func._micropath_signature is result
        mock_getsig.assert_called_once_with(func)
        mock_init.assert_called_once_with(
            func, ['a', 'b', 'c'], set(['a']), set(['b', 'c']),
            False, True,
        )

    def test_from_func_uncached_wrapper(self, mocker):
        mock_getsig = mocker.patch.object(
            injector.WantSignature, '_getsig',
            return_value=(
                ['a', 'b', 'c'], set(['a']), set(['b', 'c']),
                False, False,
            ),
        )
        mock_init = mocker.patch.object(
            injector.WantSignature, '__init__',
            return_value=None,
        )
        func = mocker.Mock(spec=[])
        wrapped = mocker.Mock(_micropath_signature=mocker.Mock(
            required=set(['b', 'd', 'e']),
            optional=set(['f']),
        ))

        result = injector.WantSignature.from_func(
            func,
            wrapped=wrapped,
            provides=['d'],
        )

        assert isinstance(result, injector.WantSignature)
        assert func._micropath_signature is result
        mock_getsig.assert_called_once_with(func)
        mock_init.assert_called_once_with(
            func, ['a', 'b', 'c'], set(['a', 'b', 'e']), set(['c', 'f']),
            False, False,
        )

    def test_init_base(self):
        result = injector.WantSignature(
            'func', ['a', 'b', 'c'], set(['a']), set(['b', 'c']),
            'pos', 'kw',
        )

        assert result.func == 'func'
        assert result.arg_order == ['a', 'b', 'c']
        assert result.required == set(['a'])
        assert result.optional == set(['b', 'c'])
        assert result.all_pos == 'pos'
        assert result.all_kw == 'kw'
        assert result.all_args == set(['a', 'b', 'c'])

    def test_init_overlap(self):
        with pytest.raises(ValueError):
            injector.WantSignature(
                'func', ['a', 'b', 'c'], set(['a', 'b']), set(['b', 'c']),
                'pos', 'kw',
            )

    def test_contains_true(self):
        obj = injector.WantSignature(
            'func', ['a', 'b', 'c'], set(['a']), set(['b', 'c']),
            False, False,
        )

        assert 'b' in obj

    def test_contains_false(self):
        obj = injector.WantSignature(
            'func', ['a', 'b', 'c'], set(['a']), set(['b', 'c']),
            False, False,
        )

        assert 'd' not in obj

    def test_contains_all_kw(self):
        obj = injector.WantSignature(
            'func', ['a', 'b', 'c'], set(['a']), set(['b', 'c']),
            False, True,
        )

        assert 'd' in obj

    def test_call_base(self, mocker):
        func = mocker.Mock()
        obj = injector.WantSignature(
            func, ['a', 'b', 'c'], set(['a']), set(['b', 'c']),
            False, False,
        )

        result = obj((), {'a': 1, 'b': 2, 'c': 3, 'd': 4, 'e': 5})

        assert result == func.return_value
        func.assert_called_once_with(a=1, b=2, c=3)

    def test_call_positional(self, mocker):
        func = mocker.Mock()
        obj = injector.WantSignature(
            func, ['self', 'a', 'b', 'c'], set(['self', 'a']), set(['b', 'c']),
            False, False,
        )

        result = obj(('self',), {'a': 1, 'b': 2, 'c': 3, 'd': 4, 'e': 5})

        assert result == func.return_value
        func.assert_called_once_with('self', a=1, b=2, c=3)

    def test_call_too_many_positional(self, mocker):
        func = mocker.Mock()
        obj = injector.WantSignature(
            func, ['a', 'b', 'c'], set(['a']), set(['b', 'c']),
            False, False,
        )

        with pytest.raises(TypeError):
            obj(
                ('a', 'b', 'c', 'd'), {'a': 1, 'b': 2, 'c': 3, 'd': 4, 'e': 5},
            )
        func.assert_not_called()

    def test_call_all_positional(self, mocker):
        func = mocker.Mock()
        obj = injector.WantSignature(
            func, ['a', 'b', 'c'], set(['a']), set(['b', 'c']),
            True, False,
        )

        result = obj(
            ('a', 'b', 'c', 'd'), {'a': 1, 'b': 2, 'c': 3, 'd': 4, 'e': 5},
        )

        assert result == func.return_value
        func.assert_called_once_with('a', 'b', 'c', 'd')

    def test_call_additional(self, mocker):
        func = mocker.Mock()
        obj = injector.WantSignature(
            func, ['a', 'b', 'c'], set(['a']), set(['b', 'c']),
            False, False,
        )

        result = obj(
            (), {'a': 1, 'b': 2, 'c': 3, 'd': 4, 'e': 5},
            {'b': 'b'},
        )

        assert result == func.return_value
        func.assert_called_once_with(a=1, b='b', c=3)

    def test_call_all_kw(self, mocker):
        func = mocker.Mock()
        obj = injector.WantSignature(
            func, ['a', 'b', 'c'], set(['a']), set(['b', 'c']),
            False, True,
        )

        result = obj(
            (), {'a': 1, 'b': 2, 'c': 3, 'd': 4, 'e': 5},
            {'b': 'b', 'f': 'f'},
        )

        assert result == func.return_value
        func.assert_called_once_with(a=1, b='b', c=3, d=4, e=5, f='f')

    def test_call_missing(self, mocker):
        func = mocker.Mock()
        obj = injector.WantSignature(
            func, ['a', 'b', 'c'], set(['a']), set(['b', 'c']),
            False, False,
        )

        with pytest.raises(TypeError):
            obj((), {'b': 2, 'c': 3, 'd': 4, 'e': 5})
        func.assert_not_called()


class TestInjectorCleanup(object):
    def test_init(self):
        result = injector.InjectorCleanup('injector')

        assert result.injector == 'injector'
        assert result.keep is None

    def test_enter(self, mocker):
        inject = mocker.Mock(_keys=set(['a', 'b', 'c']))
        obj = injector.InjectorCleanup(inject)

        result = obj.__enter__()

        assert result is inject
        assert obj.keep is not inject._keys
        assert obj.keep == inject._keys

    def test_exit(self, mocker):
        inject = mocker.MagicMock(_keys=set(['a', 'b', 'c', 'd', 'e', 'f']))
        obj = injector.InjectorCleanup(inject)
        obj.keep = set(['a', 'b', 'c'])

        result = obj.__exit__(None, None, None)

        assert result is None
        assert obj.keep is None
        inject.__delitem__.assert_has_calls([
            mocker.call('d'),
            mocker.call('e'),
            mocker.call('f'),
        ], any_order=True)
        assert inject.__delitem__.call_count == 3


class TestInjector(object):
    def test_init(self):
        result = injector.Injector()

        assert result._available == {}
        assert result._deferred == {}
        assert result._keys == set()

    def test_len(self):
        obj = injector.Injector()
        obj._keys |= set(['a', 'b', 'c'])

        assert len(obj) == 3

    def test_iter(self):
        obj = injector.Injector()
        obj._keys |= set(['a', 'b', 'c'])

        assert set(iter(obj)) == obj._keys

    def test_getitem_available(self, mocker):
        mock_call = mocker.patch.object(
            injector.Injector, '__call__',
            return_value='deferred',
        )
        obj = injector.Injector()
        obj._available['a'] = 1
        obj._keys = set(['a'])

        assert obj['a'] == 1
        assert obj._available == {'a': 1}
        mock_call.assert_not_called()

    def test_getitem_deferred(self, mocker):
        mock_call = mocker.patch.object(
            injector.Injector, '__call__',
            return_value='deferred',
        )
        obj = injector.Injector()
        obj._deferred['a'] = 'func'
        obj._keys = set(['a'])

        assert obj['a'] == 'deferred'
        assert obj._available == {'a': 'deferred'}
        mock_call.assert_called_once_with('func')

    def test_getitem_missing(self, mocker):
        mock_call = mocker.patch.object(
            injector.Injector, '__call__',
            return_value='deferred',
        )
        obj = injector.Injector()

        with pytest.raises(KeyError):
            obj['a']
        assert obj._available == {}
        mock_call.assert_not_called()

    def test_setitem(self):
        obj = injector.Injector()

        obj['a'] = 1

        assert obj._available == {'a': 1}
        assert obj._keys == set(['a'])

    def test_delitem_available_only(self):
        obj = injector.Injector()
        obj._available['a'] = 1
        obj._keys = set(['a'])

        del obj['a']

        assert obj._available == {}
        assert obj._deferred == {}
        assert obj._keys == set()

    def test_delitem_deferred_only(self):
        obj = injector.Injector()
        obj._deferred['a'] = 'deferred'
        obj._keys = set(['a'])

        del obj['a']

        assert obj._available == {}
        assert obj._deferred == {}
        assert obj._keys == set()

    def test_delitem_available_and_deferred(self):
        obj = injector.Injector()
        obj._available['a'] = 1
        obj._deferred['a'] = 'deferred'
        obj._keys = set(['a'])

        del obj['a']

        assert obj._available == {}
        assert obj._deferred == {}
        assert obj._keys == set()

    def test_delitem_missing(self):
        obj = injector.Injector()

        with pytest.raises(KeyError):
            del obj['a']
        assert obj._available == {}
        assert obj._deferred == {}
        assert obj._keys == set()

    def test_call_base(self, mocker):
        mock_ismethod = mocker.patch.object(
            injector.inspect, 'ismethod',
            return_value=False,
        )
        mock_get_method_self = mocker.patch.object(
            injector.six, 'get_method_self',
            return_value='obj',
        )
        mock_get_method_function = mocker.patch.object(
            injector.six, 'get_method_function',
            return_value='method',
        )
        mock_from_func = mocker.patch.object(
            injector.WantSignature, 'from_func',
        )
        obj = injector.Injector()

        result = obj('func', 1, 2, 3, a=4, b=5, c=6)

        assert result == mock_from_func.return_value.return_value
        mock_ismethod.assert_called_once_with('func')
        mock_get_method_self.assert_not_called()
        mock_get_method_function.assert_not_called()
        mock_from_func.assert_called_once_with('func')
        mock_from_func.return_value.assert_called_once_with(
            (1, 2, 3), obj, {'a': 4, 'b': 5, 'c': 6},
        )

    def test_call_method(self, mocker):
        mock_ismethod = mocker.patch.object(
            injector.inspect, 'ismethod',
            return_value=True,
        )
        mock_get_method_self = mocker.patch.object(
            injector.six, 'get_method_self',
            return_value='obj',
        )
        mock_get_method_function = mocker.patch.object(
            injector.six, 'get_method_function',
            return_value='method',
        )
        mock_from_func = mocker.patch.object(
            injector.WantSignature, 'from_func',
        )
        obj = injector.Injector()

        result = obj('func', 1, 2, 3, a=4, b=5, c=6)

        assert result == mock_from_func.return_value.return_value
        mock_ismethod.assert_called_once_with('func')
        mock_get_method_self.assert_called_once_with('func')
        mock_get_method_function.assert_called_once_with('func')
        mock_from_func.assert_called_once_with('method')
        mock_from_func.return_value.assert_called_once_with(
            ('obj', 1, 2, 3), obj, {'a': 4, 'b': 5, 'c': 6},
        )

    def test_call_no_func(self, mocker):
        mock_ismethod = mocker.patch.object(
            injector.inspect, 'ismethod',
            return_value=False,
        )
        mock_get_method_self = mocker.patch.object(
            injector.six, 'get_method_self',
            return_value='obj',
        )
        mock_get_method_function = mocker.patch.object(
            injector.six, 'get_method_function',
            return_value='method',
        )
        mock_from_func = mocker.patch.object(
            injector.WantSignature, 'from_func',
        )
        obj = injector.Injector()

        with pytest.raises(TypeError):
            obj(a=4, b=5, c=6)
        mock_ismethod.assert_not_called()
        mock_get_method_self.assert_not_called()
        mock_get_method_function.assert_not_called()
        mock_from_func.assert_not_called()
        mock_from_func.return_value.assert_not_called()

    def test_set_deferred(self):
        obj = injector.Injector()

        obj.set_deferred('a', 'deferred')

        assert obj._deferred == {'a': 'deferred'}

    def test_cleanup(self, mocker):
        mock_InjectorCleanup = mocker.patch.object(
            injector, 'InjectorCleanup',
        )
        obj = injector.Injector()

        result = obj.cleanup()

        assert result == mock_InjectorCleanup.return_value
        mock_InjectorCleanup.assert_called_once_with(obj)


class TestInject(object):
    def test_base(self, mocker):
        mock_from_func = mocker.patch.object(
            injector.WantSignature, 'from_func',
        )
        func = mocker.Mock()

        decorator = injector.inject()

        assert callable(decorator)
        mock_from_func.assert_not_called()

        result = decorator(func)

        assert result == func
        mock_from_func.assert_called_once_with(
            func,
            required=None,
            optional=None,
        )
        func.assert_not_called()

    def test_alt(self, mocker):
        mock_from_func = mocker.patch.object(
            injector.WantSignature, 'from_func',
        )
        func = mocker.Mock()

        decorator = injector.inject(required='required', optional='optional')

        assert callable(decorator)
        mock_from_func.assert_not_called()

        result = decorator(func)

        assert result == func
        mock_from_func.assert_called_once_with(
            func,
            required='required',
            optional='optional',
        )
        func.assert_not_called()


class TestWraps(object):
    def test_base(self, mocker):
        mock_from_func = mocker.patch.object(
            injector.WantSignature, 'from_func',
        )
        mock_wraps = mocker.patch.object(injector.six, 'wraps')
        func = mocker.Mock()

        decorator = injector.wraps('wrapped')

        assert callable(decorator)
        mock_from_func.assert_not_called()
        mock_wraps.assert_not_called()

        result = decorator(func)

        assert result == mock_wraps.return_value.return_value
        assert result._micropath_signature == mock_from_func.return_value
        mock_from_func.assert_called_once_with(
            func, 'wrapped', None, [], [],
        )
        mock_wraps.assert_called_once_with(
            'wrapped', functools.WRAPPER_ASSIGNMENTS,
            functools.WRAPPER_UPDATES,
        )
        mock_wraps.return_value.assert_called_once_with(func)

    def test_alt(self, mocker):
        mock_from_func = mocker.patch.object(
            injector.WantSignature, 'from_func',
        )
        mock_wraps = mocker.patch.object(injector.six, 'wraps')
        func = mocker.Mock()

        decorator = injector.wraps(
            'wrapped', 'assigned', 'updated', 'provides', 'required',
            'optional',
        )

        assert callable(decorator)
        mock_from_func.assert_not_called()
        mock_wraps.assert_not_called()

        result = decorator(func)

        assert result == mock_wraps.return_value.return_value
        assert result._micropath_signature == mock_from_func.return_value
        mock_from_func.assert_called_once_with(
            func, 'wrapped', 'provides', 'required', 'optional',
        )
        mock_wraps.assert_called_once_with(
            'wrapped', 'assigned', 'updated',
        )
        mock_wraps.return_value.assert_called_once_with(func)


class TestCallWrapped(object):
    def test_base(self, mocker):
        mock_from_func = mocker.patch.object(
            injector.WantSignature, 'from_func',
        )

        result = injector.call_wrapped('func', 'args', 'kwargs')

        assert result == mock_from_func.return_value.return_value
        mock_from_func.assert_called_once_with('func')
        mock_from_func.return_value.assert_called_once_with('args', 'kwargs')


class TestWants(object):
    def test_true(self, mocker):
        mock_from_func = mocker.patch.object(
            injector.WantSignature, 'from_func',
            return_value=set(['a']),
        )

        result = injector.wants('func', 'a')

        assert result is True
        mock_from_func.assert_called_once_with('func')

    def test_false(self, mocker):
        mock_from_func = mocker.patch.object(
            injector.WantSignature, 'from_func',
            return_value=set(['a']),
        )

        result = injector.wants('func', 'b')

        assert result is False
        mock_from_func.assert_called_once_with('func')



# Copyright (C) 2018 by Kevin L. Mitchell <klmitch@mit.edu>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you
# may not use this file except in compliance with the License. You may
# obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied. See the License for the specific language governing
# permissions and limitations under the License.

from micropath.controller import Controller
from micropath.elements import (
    SkipBinding, Delegation, path, bind, route, mount,
)
from micropath.injector import inject, wraps, call_wrapped, wants
from micropath.request import Request

__all__ = [
    'Controller',

    'SkipBinding', 'Delegation', 'path', 'bind', 'route', 'mount',

    'inject', 'wraps', 'call_wrapped', 'wants',

    'Request',
]



# Copyright (C) 2018 by Kevin L. Mitchell <klmitch@mit.edu>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you
# may not use this file except in compliance with the License. You may
# obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied. See the License for the specific language governing
# permissions and limitations under the License.

import sys
import traceback
from wsgiref import simple_server

import six
import webob.dec
import webob.exc

from micropath import elements
from micropath import injector
from micropath import request


class ControllerMeta(type):
    """
    Metaclass for controllers.  This metaclass takes care of the task
    of collecting routing elements defined in the class's namespace
    and putting them into a single routing tree.
    """

    def __new__(cls, name, bases, namespace):
        """
        Construct a new ``Controller`` subclass.

        :param str name: The name of the new class.
        :param bases: A tuple of base classes.
        :type bases: ``tuple`` of classes
        :param dict namespace: The namespace of the new class.

        :returns: A newly constructed subclass of the ``Controller``
                  class.
        """

        # Construct a new Root
        root = elements.Root()

        # Collect delegations
        delegations = {}

        # Walk the namespace looking for elements and callables with
        # Method elements
        for ident, value in namespace.items():
            # Add elements to the root
            if isinstance(value, elements.Element):
                root.add_elem(value, ident)
                continue

            # Add HTTP methods to the root as well
            if hasattr(value, '_micropath_methods'):
                for meth in value._micropath_methods:
                    root.add_elem(meth, ident)

            # Mount the delegation to the root
            if isinstance(value, elements.Delegation):
                if value.element is None:
                    root.mount(value)
                else:
                    root.add_elem(value.element, ident)

                # Add it to the set of delegations
                delegations[ident] = value
                continue

            # Deal with handlers
            if getattr(value, '_micropath_handler', False):
                # If it doesn't have an elem, set it to the root
                if getattr(value, '_micropath_elem', None) is None:
                    value._micropath_elem = root

        # Add the root and handlers list to the namespace
        namespace.update(
            _micropath_root=root,
            _micropath_delegations=[
                delegation for _name, delegation in
                sorted(delegations.items(), key=lambda x: x[0])
            ],
        )

        return super(ControllerMeta, cls).__new__(cls, name, bases, namespace)


@six.add_metaclass(ControllerMeta)
class Controller(object):
    """
    Controller class.  This is the central type for routing URL paths
    and HTTP methods to actual handlers.  See the
    ``micropath.path()``, ``micropath.bind()``, and
    ``micropath.mount()`` functions and the ``@micropath.route()``
    decorator for how to create the necessary routes.

    This class has some attributes that subclasses may set if desired,
    to control behavior.  Those class attributes are as follows:

    * ``micropath_request`` - The class to use for representing a
        request.  This must be a subclass of ``micropath.Request``.
    * ``micropath_methods`` - The ``micropath`` framework has a
        default implementation for the "OPTIONS" HTTP method.  This
        class attribute should contain a set of the HTTP methods that
        are recognized if a default route is available (one for which
        no methods have been passed to the ``@micropath.route()``
        decorator).  The default contains "HEAD", "GET", "PUT",
        "POST", "DELETE", and of course "OPTIONS".
    * ``micropath_debug`` - A class attribute that is set to
        ``False``.  This may be shadowed by an instance attribute of
        the same name to enable debugging responses; this will include
        traceback and debugging information in the 500 error generated
        if the ``micropath_server_error()`` hook method raises an
        exception.  The default implementations of
        ``micropath_request_error()`` and ``micropath_server_error()``
        also include traceback and debugging information if this
        attribute is ``True``.  WARNING: THIS ATTRIBUTE MUST NOT BE
        SET TO ``True`` ON PRODUCTION SERVERS!  Exception information
        can leak security-sensitive data to callers.
    * ``micropath_request_attrs`` - A dictionary of request attributes
        that may be injected into handler methods.  The keys of this
        dictionary are the parameter names the handler methods may
        request, and the values are the names of the attribute of the
        request.  A value of ``None`` indicates that the attribute has
        the same name as the parameter name.  The default value allows
        all request attributes to be injected; subclasses that
        override this attribute may wish to copy the value from
        ``micropath.Controller`` and then update it to list additional
        attributes.

    In addition to the class attributes listed above, subclasses may
    also override several methods to control behavior.  The methods
    that may be overridden are as follows:

    * ``micropath_construct()`` - Called when a mounted class is
        traversed.  The default implementation constructs the mounted
        class with only the passed keyword arguments, so if using
        ``micropath.mount()`` and the controller classes require other
        arguments (e.g., configuration), the class must implement this
        method.  Since ``micropath`` does not provide any constraints
        on your controller class's ``__init__()`` method, the
        intention of this hook method is to allow the mounted class to
        be properly configured.  For instance, if you load a
        configuration file and then pass it to the ``config``
        parameter of your root controller class's ``__init__()``
        method, you should implement ``micropath_construct()`` to
        construct the other class with that same configuration.
    * ``micropath_prepare_injector()`` - Called immediately after the
        dependency injector has been initialized.  This can be used to
        add additional fields to the dependency injector.  Only the
        implementation in the root ``Controller`` subclass is useful.
    * ``micropath_server_error()`` - Called if a handler method or the
        injector raises any exception that is not
        ``webob.exc.HTTPException``.  The default implementation
        returns a bare ``webob.exc.HTTPInternalServerError``, which
        causes a 500 error to be returned to the HTTP client.  This
        must be implemented in each ``Controller`` subclass, so if
        customizing this hook, it is recommended to create a base
        class with the desired implementation, then subclass that.
    * ``micropath_request_error()`` - Called if a handler method
        requests an attribute of the request object, but an error
        occurs while accessing that attribute.  This typically
        indicates a format error with the requested resource, such as
        invalid JSON data when attempting to access ``json_body``.  As
        such, the default implementation returns a bare
        ``webob.exc.HTTPBadRequest``, which causes a 400 error to be
        returned to the HTTP client.  Only the implementation in the
        root ``Controller`` subclass is useful.  Also note that the
        return value of this hook method must be a subclass of
        ``Exception``.
    * ``micropath_not_found()`` - Called if the path could not be
        resolved.  Note that this will not be called for a given
        request if the handler function for that URL and HTTP method
        includes the ``path_info`` parameter.  This must be
        implemented in each ``Controller`` subclass, so if customizing
        this hook, it is recommended to create a base class with the
        desired implementation, then subclass that.
    * ``micropath_not_implemented()`` - Called if the HTTP method is
        not implemented.  This must be implemented in each
        ``Controller`` subclass, so if customizing this hook, it is
        recommended to create a base class with the desired
        implementation, then subclass that.
    * ``micropath_options()`` - Called if the HTTP method is "OPTIONS"
        and the method isn't routed to a handler.  This method must
        create a response that includes the "Allow" header with a
        comma-separated list of methods.  This must be implemented in
        each ``Controller`` subclass, so if customizing this hook, it
        is recommended to create a base class with the desired
        implementation, then subclass that.

    The ``Controller`` base class also defines an ``__init__()``
    method that all subclasses must ensure is called with no
    arguments.  This base ``__init__()`` method instantiates the
    mounted controllers, in order to ensure that there are no race
    conditions during request processing in the face of threaded WSGI
    servers.  This means that the ``micropath`` framework is itself
    thread-safe, without needing to use any threading primitives.

    Finally, a convenience ``micropath_run()`` method is provided for
    testing purposes (it should not be used on production systems).
    This method serves the WSGI application at a specified host and
    port.
    """

    # Set the default class to use for requests
    micropath_request = request.Request

    # The set of methods to be considered implemented if a default
    # route was added to a handler
    micropath_methods = set([
        'HEAD', 'GET', 'PUT', 'POST', 'DELETE', 'OPTIONS'
    ])

    # Controls debugging.  THIS ATTRIBUTE MUST NOT BE True ON
    # PRODUCTION SERVERS!
    micropath_debug = False

    # Set the default set of request attributes to make injectable
    micropath_request_attrs = {
        'get': 'GET',
        'post': 'POST',
        'accept': None,
        'accept_charset': None,
        'accept_encoding': None,
        'accept_language': None,
        'application_url': None,
        'authorization': None,
        'base_path': None,
        'body': None,
        'body_file': None,
        'body_file_raw': None,
        'body_file_seekable': None,
        'cache_control': None,
        'charset': None,
        'client_addr': None,
        'content_length': None,
        'content_type': None,
        'cookies': None,
        'date': None,
        'domain': None,
        'environ': None,
        'headers': None,
        'host': None,
        'host_port': None,
        'host_url': None,
        'http_version': None,
        'if_match': None,
        'if_modified_since': None,
        'if_none_match': None,
        'if_range': None,
        'if_unmodified_since': None,
        'injector': None,
        'is_body_readable': None,
        'is_body_seekable': None,
        'is_xhr': None,
        'json': None,
        'json_body': None,
        'max_forwards': None,
        'method': None,
        'params': None,
        'path': None,
        'path_info': None,
        'path_qs': None,
        'path_url': None,
        'pragma': None,
        'query_string': None,
        'range': None,
        'referer': None,
        'referrer': None,
        'remote_addr': None,
        'remote_user': None,
        'scheme': None,
        'script_name': None,
        'server_name': None,
        'server_port': None,
        'text': None,
        'url': None,
        'url_encoding': None,
        'urlargs': None,
        'urlvars': None,
        'user_agent': None,
    }

    def __init__(self):
        """
        Initialize a ``Controller`` instance.  Subclasses must call the
        superclass method to properly initialize the controller.
        """

        # Initialize delegations; this ensures that all subordinate
        # controllers are initialized at the time the root controller
        # is, thus avoiding any threading issues within the
        # Delegations.get() logic
        for deleg in self._micropath_delegations:
            deleg.get(self)

    def __call__(self, environ, start_response):
        """
        Contains the implementation of the WSGI application which is the
        heart of the ``micropath`` framework.  This allows instances
        of subclasses of ``Controller`` to be used as WSGI
        applications.

        :param dict environ: The WSGI environment.
        :param start_response: The callable used to start the
                               response.

        :returns: The return value of invoking the response WSGI
                  application, as returned by the handler methods.  If
                  no handler method is available, an appropriate error
                  will be generated to indicate that the path doesn't
                  exist (404) or the method is unacceptable (501).
        """

        # Note: The contents of this method are mostly copied from the
        # __call__() method of webob.dec.wsgify

        # First, construct the request and set the default response
        req = self.micropath_request(environ)
        req.response = req.ResponseClass()

        # Next, walk the path tree and invoke the handler; we use the
        # injector cleanup context manager to explicitly break
        # reference loops after we've dispatched to the handler method
        with req.injector.cleanup() as injector:
            try:
                # Add anything from the request's urlvars
                injector.update(req.urlvars)

                # Populate the request and root_controller fields of
                # the injector
                injector['request'] = req
                injector['root_controller'] = self

                # Add deferred accessors for all the other fields
                def defer(key):
                    # Can't use operator.attrgetter because we need
                    # the parameter to be named "request"
                    def get(request):
                        try:
                            return getattr(request, key)
                        except Exception:
                            raise self.micropath_request_error(
                                request, key, sys.exc_info(),
                            )
                    return get
                for key, mapped in self.micropath_request_attrs.items():
                    injector.set_deferred(key, defer(mapped or key))

                # Hook for setting up additional injection settings
                self.micropath_prepare_injector(req, injector)

                resp = self._micropath_dispatch(req, injector)
            except webob.exc.HTTPException as exc:
                resp = exc
            except Exception:
                # Some other error occurred; we'll turn it into an
                # internal server error
                try:
                    resp = self.micropath_server_error(req, sys.exc_info())
                except Exception:
                    # An exception could be raised by
                    # micropath_server_error(); this clause acts as an
                    # absolute last resort to ensure that no exception
                    # makes it all the way up the stack.  Begin by
                    # formulating the detail for debugging purposes
                    detail = None
                    if self.micropath_debug:
                        # Debugging detail requested; format traceback
                        detail = traceback.format_exc()

                    # Construct the exception
                    resp = webob.exc.HTTPInternalServerError(detail)

        # Use the default response if none was returned
        if resp is None:
            resp = req.response

        # Convert text and bytes
        if isinstance(resp, six.text_type):
            resp = resp.encode(req.charset)
        if isinstance(resp, bytes):
            body = resp
            resp = req.response
            resp.write(body)

        # Merge the cookies
        if resp is not req.response:
            resp = req.response.merge_cookies(resp)

        # Return the response
        return resp(environ, start_response)

    def _micropath_dispatch(self, req, inj):
        """
        Walk the element tree based on the URL path in the request.  This
        method is used to locate the correct handler to invoke for a
        given request, or to invoke a proper delegation to handle the
        next part of the request URL.

        :param req: The request being processed.
        :type req: ``micropath.Request``
        :param inj: The injector.  This will have bindings added to
                    it, and will be used to invoke the handler method.
        :type inj: ``micropath.injector.Injector``

        :returns: The result of invoking the handler method.
        """

        # Resolve the path to a Path or Binding
        elem, path_info_required = self._micropath_resolve(req, inj)

        # Get the matching function and delegation
        func, delegation = self._micropath_delegation(req, elem)

        # First, is there a function?
        if (func and
                (not path_info_required or injector.wants(func, 'path_info'))):
            return inj(func, self)

        # OK, delegate if needed
        if delegation:
            # Get the object being delegated to
            obj = delegation.get(self)

            # Dispatch to it
            return obj._micropath_dispatch(req, inj)

        # We couldn't find an implementation...
        if path_info_required or not elem.methods:
            # Couldn't find the path; we'll return a 404
            return self.micropath_not_found(req, req.path_info)
        elif elem.methods and req.method == 'OPTIONS':
            meths = self._micropath_methods(elem)

            # Generate and return the OPTIONS response
            return self.micropath_options(req, list(sorted(meths)))

        # Couldn't find an implementation for the method
        return self.micropath_not_implemented(req, req.method)

    def _micropath_resolve(self, req, inj):
        """
        Resolve the request URL path to a specific path element.  This
        will set up injections for any variable bindings, as well as
        setting the variable bindings on the request's ``urlvars``
        dictionary.

        :param req: The request being processed.
        :type req: ``micropath.Request``
        :param inj: The injector.  This will have bindings added to
                    it.
        :type inj: ``micropath.injector.Injector``

        :returns: A tuple of the found path element (either
                  ``micropath.elements.Path`` or
                  ``micropath.elements.Binding``) and a boolean
                  indicating whether the ``path_info`` of the request
                  has been exhausted--if ``True``, the implementing
                  function must want the ``path_info`` attribute, or a
                  404 will be generated.
        """

        # Must the handler have path_info?
        path_info_required = True

        # Start at the controller's root
        elem = self._micropath_root
        path_elem = req.path_info_peek()
        while path_elem:
            # If it's a static path, we'll go down that branch
            if path_elem in elem.paths:
                elem = elem.paths[path_elem]
            elif elem.bindings is not None:
                try:
                    value = elem.bindings.validate(self, inj, path_elem)
                except elements.SkipBinding:
                    break

                # Save it into the injector and the urlvars
                if elem.bindings.ident not in inj:
                    # Only add to the injector if there are no
                    # conflicts
                    inj[elem.bindings.ident] = value
                req.urlvars[elem.bindings.ident] = value

                # Set the next element
                elem = elem.bindings
            else:
                break

            # OK, pop off the path element and set up for the next one
            req.path_info_pop()
            path_elem = req.path_info_peek()
        else:
            # Ran off the end of the path_info
            path_info_required = False

        return elem, path_info_required

    def _micropath_delegation(self, req, elem):
        """
        Determine what function or mount point to delegate the request to.

        :param req: The request being processed.
        :type req: ``micropath.Request``

        :param elem: The path element at which to resolve the
                     delegation.
        :type elem: ``micropath.elements.Element``

        :returns: A tuple of the function and a delegation.  Either or
                  both may be ``None``.  If the delegation is not
                  ``None``, it will be an instance of
                  ``micropath.elements.Delegation``.
        """

        # Pick out the Method instance
        meth = elem.methods.get(
            'GET' if req.method == 'HEAD' else req.method,
            elem.methods.get(None),
        )

        return (
            meth.func if meth else None,
            (meth.delegation or elem.delegation) if meth else elem.delegation,
        )

    def _micropath_methods(self, elem):
        """
        Determine the methods available at a particular point in the URL
        hierarchy.

        :param elem: The path element.
        :type elem: ``micropath.elements.Element``

        :returns: A set of the HTTP methods that are implemented at
                  this point of the URL hierarchy.
        :rtype: ``set`` of ``str``
        """

        # First, get the set of methods we see in elem.methods
        meths = set(m for m in elem.methods if m is not None)

        # Next, add the micropath_methods if None is in
        # elem.methods
        if None in elem.methods:
            meths |= self.micropath_methods

        # Now, add the fixed options: HEAD (if GET is present)
        # and, of course, OPTIONS
        if 'GET' in meths:
            meths.add('HEAD')
        meths.add('OPTIONS')

        return meths

    def micropath_construct(self, other, kwargs):
        """
        Construct another controller.  This is called on objects which
        have other controller classes mounted on them, in order to
        construct the other controller.

        :param other: The other controller class to construct.
        :param dict kwargs: Additional keyword arguments passed when
                            setting up the mount point.

        :returns: An instance of the other controller class, properly
                  initialized.
        """

        return other(**kwargs)

    def micropath_prepare_injector(self, request, injector):
        """
        A hook method for injector preparation.  This allows subclasses to
        add things to the injector if needed.

        :param request: The request being processed.
        :type request: ``micropath.Request``
        :param injector: The injector being prepared.  Implementations
                         can assign values using dictionary syntax, or
                         may set functions that will be called with
                         appropriate arguments (using dependency
                         injection) using the
                         ``injector.set_deferred()`` method.  The
                         ``injector.set_deferred()`` method takes the
                         name of the parameter and the implementing
                         function as its two required arguments, and
                         the function is expected to return the value
                         that will be injected as an argument.
        :type injector: ``micropath.injector.Injector``
        """

        pass  # pragma: no cover

    def micropath_server_error(self, request, cause):
        """
        A hook method for handling the case that an exception that was not
        a ``webob.exc.HTTPException`` was raised during request
        processing.

        :param request: The request that caused the error to be
                        raised.
        :param cause: The exception tuple for the exception that
                      caused this method to be called.  This will be
                      the tuple from ``sys.exc_info()``.

        :returns: An appropriate response.  The default implementation
                  returns a bare
                  ``webob.exc.HTTPInternalServerError``, including
                  traceback information only if ``micropath_debug`` is
                  set to ``True``.  NOTE THAT, FOR SECURITY REASONS,
                  THE TRACEBACK SHOULD *NOT* BE INCLUDED ON PRODUCTION
                  SYSTEMS: THIS INFORMATION MAY INCLUDE SENSITIVE
                  DATA, SUCH AS PATHS.
        """

        # Formulate the detail for debugging purposes
        detail = None
        if self.micropath_debug:
            # Debugging detail requested; format traceback
            detail = ''.join(traceback.format_exception(*cause))

        # Construct and return the exception
        return webob.exc.HTTPInternalServerError(detail)

    def micropath_request_error(self, request, attr, cause):
        """
        A hook method for handling the case that an exception that was not
        a ``webob.exc.HTTPException`` was raised while injecting a
        request attribute.  This might occur if a JSON payload was not
        valid JSON, for instance.

        :param request: The request that caused the error to be
                        raised.
        :param str attr: The name of the attribute that could not be
                         accessed.
        :param cause: The exception tuple for the exception that
                      caused this method to be called.  This will be
                      the tuple from ``sys.exc_info()``.

        :returns: An appropriate response.  The default implementation
                  returns a bare ``webob.exc.HTTPBadRequest``,
                  including traceback information only if
                  ``micropath_debug`` is set to ``True``.  NOTE THAT,
                  FOR SECURITY REASONS, THE TRACEBACK SHOULD *NOT* BE
                  INCLUDED ON PRODUCTION SYSTEMS: THIS INFORMATION MAY
                  INCLUDE SENSITIVE DATA, SUCH AS PATHS.  Also note
                  that the return value of this function MUST be an
                  object extending ``Exception``, such as
                  ``webob.exc.HTTPBadRequest``; the value will be
                  passed to ``raise``.
        """

        # Formulate the detail for debugging purposes
        detail = None
        if self.micropath_debug:
            # Debugging detail requested; format traceback
            detail = 'Accessing request attribute "%s":\n%s' % (
                attr, ''.join(traceback.format_exception(*cause)),
            )

        # Construct and return the exception
        return webob.exc.HTTPBadRequest(detail)

    def micropath_not_found(self, request, path_info):
        """
        A hook method for the case that the path could not be fully
        resolved.

        :param request: The request that caused the error to be
                        encountered.
        :param str path_info: The remaining portions of the URL path
                              that could not be resolved.

        :returns: An appropriate response.  The default implementation
                  returns a bare ``webob.exc.HTTPNotfound``.
        """

        return webob.exc.HTTPNotFound()

    def micropath_not_implemented(self, request, method):
        """
        A hook method for the case that the request method is not
        implemented.

        :param request: The request that caused the error to be
                        encountered.
        :param str method: The HTTP method that was not implemented.

        :returns: An appropriate response.  The default implementation
                  returns a bare ``webob.exc.HTTPNotImplemented``.
        """

        return webob.exc.HTTPNotImplemented()

    def micropath_options(self, request, methods):
        """
        A hook method for implementing the response to the OPTIONS HTTP
        method.

        :param request: The request that caused the default OPTIONS
                        handler to be invoked.
        :param methods: A sorted list of HTTP methods.  This should be
                        turned into a comma-separated list and placed
                        in the response's "Allow" header.

        :returns: An appropriate response.  The default implementation
                  returns a ``webob.exc.HTTPNoContent`` with the
                  "Allow" header set to the comma-separated list of
                  allowed HTTP methods.
        """

        return webob.exc.HTTPNoContent(headers={'Allow': ','.join(methods)})

    def micropath_run(self, host='127.0.0.1', port=8000):
        """
        Serve the application.  This makes use of the Python standard
        library ``wsgiref`` package's ``simple_server`` to present the
        application defined in this controller as a WSGI server.

        Note: It is NOT RECOMMENDED to use this method to serve an
        application for production purposes.  No attempt is made to
        handle threading, error handling, SSL, or a host of other
        issues that real, production-ready HTTP and WSGI servers
        handle.  This method is provided for testing purposes only.

        :param str host: The IP address for the server to listen on.
                         Defaults to "127.0.0.1", meaning to listen on
                         the loopback interface (aka, "localhost".)
        :param int port: The port for the server to listen on.
                         Defaults to 8000.
        """

        server = simple_server.make_server(host, port, self)
        server.serve_forever()



# Copyright (C) 2018 by Kevin L. Mitchell <klmitch@mit.edu>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you
# may not use this file except in compliance with the License. You may
# obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied. See the License for the specific language governing
# permissions and limitations under the License.

import pytest

from micropath import elements
from micropath import request


class TestRequest(object):
    def test_init(self, mocker):
        result = request.Request.blank(
            '/is/a/test',
            base_url='http://example.com/this',
        )

        assert result.environ['micropath.base_path'] == '/this'

    def test_url_for_base(self, mocker):
        elems = {
            'a': mocker.Mock(spec=elements.Path, t_parent=None),
            'b': mocker.Mock(spec=elements.Binding, t_parent='a', **{
                'format.return_value': '1',
            }),
            'c': mocker.Mock(spec=elements.Path, t_parent=None),
            'd': mocker.Mock(spec=elements.Binding, t_parent='c', **{
                'format.return_value': '2',
            }),
        }
        for ident, elem in elems.items():
            elem.ident = ident
            elem.parent = elems[elem.t_parent] if elem.t_parent else None
        controllers = [
            mocker.Mock(root='a', elem=None),
            mocker.Mock(root='c', elem='b'),
        ]
        parent = None
        for cont in controllers:
            cont._micropath_parent = parent
            cont._micropath_elem = (
                elems[cont.elem] if cont.elem else None
            )
            root = mocker.Mock(spec=elements.Root)
            root.parent = None
            elems[cont.root].parent = root
            parent = cont
        meth = mocker.Mock(_micropath_elem=elems['d'])
        mock_get_method_self = mocker.patch.object(
            request.six, 'get_method_self',
            return_value=controllers[-1],
        )
        mock_isclass = mocker.patch.object(
            request.inspect, 'isclass',
            return_value=False,
        )
        obj = request.Request.blank(
            '/is/a/test',
            base_url='http://example.com/this',
        )

        result = obj.url_for(meth, b=1, d=2)

        assert result == 'http://example.com/this/a/1/c/2'
        mock_get_method_self.assert_called_once_with(meth)
        mock_isclass.assert_called_once_with(controllers[-1])

    def test_url_for_too_few_arguments(self, mocker):
        elems = {
            'a': mocker.Mock(spec=elements.Path, t_parent=None),
            'b': mocker.Mock(spec=elements.Binding, t_parent='a', **{
                'format.return_value': '1',
            }),
            'c': mocker.Mock(spec=elements.Path, t_parent=None),
            'd': mocker.Mock(spec=elements.Binding, t_parent='c', **{
                'format.return_value': '2',
            }),
        }
        for ident, elem in elems.items():
            elem.ident = ident
            elem.parent = elems[elem.t_parent] if elem.t_parent else None
        controllers = [
            mocker.Mock(root='a', elem=None),
            mocker.Mock(root='c', elem='b'),
        ]
        parent = None
        for cont in controllers:
            cont._micropath_parent = parent
            cont._micropath_elem = (
                elems[cont.elem] if cont.elem else None
            )
            root = mocker.Mock(spec=elements.Root)
            root.parent = None
            elems[cont.root].parent = root
            parent = cont
        mock_get_method_self = mocker.patch.object(
            request.six, 'get_method_self',
            return_value=controllers[-1],
        )
        mock_isclass = mocker.patch.object(
            request.inspect, 'isclass',
            return_value=False,
        )
        obj = request.Request.blank(
            '/is/a/test',
            base_url='http://example.com/this',
        )

        with pytest.raises(TypeError):
            obj.url_for(b=1, d=2)
        mock_get_method_self.assert_not_called()
        mock_isclass.assert_not_called()

    def test_url_for_too_many_arguments(self, mocker):
        elems = {
            'a': mocker.Mock(spec=elements.Path, t_parent=None),
            'b': mocker.Mock(spec=elements.Binding, t_parent='a', **{
                'format.return_value': '1',
            }),
            'c': mocker.Mock(spec=elements.Path, t_parent=None),
            'd': mocker.Mock(spec=elements.Binding, t_parent='c', **{
                'format.return_value': '2',
            }),
        }
        for ident, elem in elems.items():
            elem.ident = ident
            elem.parent = elems[elem.t_parent] if elem.t_parent else None
        controllers = [
            mocker.Mock(root='a', elem=None),
            mocker.Mock(root='c', elem='b'),
        ]
        parent = None
        for cont in controllers:
            cont._micropath_parent = parent
            cont._micropath_elem = (
                elems[cont.elem] if cont.elem else None
            )
            root = mocker.Mock(spec=elements.Root)
            root.parent = None
            elems[cont.root].parent = root
            parent = cont
        meth = mocker.Mock(_micropath_elem=elems['d'])
        mock_get_method_self = mocker.patch.object(
            request.six, 'get_method_self',
            return_value=controllers[-1],
        )
        mock_isclass = mocker.patch.object(
            request.inspect, 'isclass',
            return_value=False,
        )
        obj = request.Request.blank(
            '/is/a/test',
            base_url='http://example.com/this',
        )

        with pytest.raises(TypeError):
            obj.url_for(meth, 'too', 'many', b=1, d=2)
        mock_get_method_self.assert_not_called()
        mock_isclass.assert_not_called()

    def test_url_for_non_callable(self, mocker):
        elems = {
            'a': mocker.Mock(spec=elements.Path, t_parent=None),
            'b': mocker.Mock(spec=elements.Binding, t_parent='a', **{
                'format.return_value': '1',
            }),
            'c': mocker.Mock(spec=elements.Path, t_parent=None),
            'd': mocker.Mock(spec=elements.Binding, t_parent='c', **{
                'format.return_value': '2',
            }),
        }
        for ident, elem in elems.items():
            elem.ident = ident
            elem.parent = elems[elem.t_parent] if elem.t_parent else None
        controllers = [
            mocker.Mock(root='a', elem=None),
            mocker.Mock(root='c', elem='b'),
        ]
        parent = None
        for cont in controllers:
            cont._micropath_parent = parent
            cont._micropath_elem = (
                elems[cont.elem] if cont.elem else None
            )
            root = mocker.Mock(spec=elements.Root)
            root.parent = None
            elems[cont.root].parent = root
            parent = cont
        meth = mocker.NonCallableMock(_micropath_elem=elems['d'])
        mock_get_method_self = mocker.patch.object(
            request.six, 'get_method_self',
            return_value=controllers[-1],
        )
        mock_isclass = mocker.patch.object(
            request.inspect, 'isclass',
            return_value=False,
        )
        obj = request.Request.blank(
            '/is/a/test',
            base_url='http://example.com/this',
        )

        with pytest.raises(ValueError):
            obj.url_for(meth, b=1, d=2)
        mock_get_method_self.assert_not_called()
        mock_isclass.assert_not_called()

    def test_url_for_no_element(self, mocker):
        elems = {
            'a': mocker.Mock(spec=elements.Path, t_parent=None),
            'b': mocker.Mock(spec=elements.Binding, t_parent='a', **{
                'format.return_value': '1',
            }),
            'c': mocker.Mock(spec=elements.Path, t_parent=None),
            'd': mocker.Mock(spec=elements.Binding, t_parent='c', **{
                'format.return_value': '2',
            }),
        }
        for ident, elem in elems.items():
            elem.ident = ident
            elem.parent = elems[elem.t_parent] if elem.t_parent else None
        controllers = [
            mocker.Mock(root='a', elem=None),
            mocker.Mock(root='c', elem='b'),
        ]
        parent = None
        for cont in controllers:
            cont._micropath_parent = parent
            cont._micropath_elem = (
                elems[cont.elem] if cont.elem else None
            )
            root = mocker.Mock(spec=elements.Root)
            root.parent = None
            elems[cont.root].parent = root
            parent = cont
        meth = mocker.Mock(_micropath_elem=None)
        mock_get_method_self = mocker.patch.object(
            request.six, 'get_method_self',
            return_value=controllers[-1],
        )
        mock_isclass = mocker.patch.object(
            request.inspect, 'isclass',
            return_value=False,
        )
        obj = request.Request.blank(
            '/is/a/test',
            base_url='http://example.com/this',
        )

        with pytest.raises(ValueError):
            obj.url_for(meth, b=1, d=2)
        mock_get_method_self.assert_not_called()
        mock_isclass.assert_not_called()

    def test_url_for_class_method(self, mocker):
        elems = {
            'a': mocker.Mock(spec=elements.Path, t_parent=None),
            'b': mocker.Mock(spec=elements.Binding, t_parent='a', **{
                'format.return_value': '1',
            }),
            'c': mocker.Mock(spec=elements.Path, t_parent=None),
            'd': mocker.Mock(spec=elements.Binding, t_parent='c', **{
                'format.return_value': '2',
            }),
        }
        for ident, elem in elems.items():
            elem.ident = ident
            elem.parent = elems[elem.t_parent] if elem.t_parent else None
        controllers = [
            mocker.Mock(root='a', elem=None),
            mocker.Mock(root='c', elem='b'),
        ]
        parent = None
        for cont in controllers:
            cont._micropath_parent = parent
            cont._micropath_elem = (
                elems[cont.elem] if cont.elem else None
            )
            root = mocker.Mock(spec=elements.Root)
            root.parent = None
            elems[cont.root].parent = root
            parent = cont
        meth = mocker.Mock(_micropath_elem=elems['d'])
        mock_get_method_self = mocker.patch.object(
            request.six, 'get_method_self',
            return_value=controllers[-1],
        )
        mock_isclass = mocker.patch.object(
            request.inspect, 'isclass',
            return_value=True,
        )
        obj = request.Request.blank(
            '/is/a/test',
            base_url='http://example.com/this',
        )

        with pytest.raises(ValueError):
            obj.url_for(meth, b=1, d=2)
        mock_get_method_self.assert_called_once_with(meth)
        mock_isclass.assert_called_once_with(controllers[-1])

    def test_url_for_missing_binding(self, mocker):
        elems = {
            'a': mocker.Mock(spec=elements.Path, t_parent=None),
            'b': mocker.Mock(spec=elements.Binding, t_parent='a', **{
                'format.return_value': '1',
            }),
            'c': mocker.Mock(spec=elements.Path, t_parent=None),
            'd': mocker.Mock(spec=elements.Binding, t_parent='c', **{
                'format.return_value': '2',
            }),
        }
        for ident, elem in elems.items():
            elem.ident = ident
            elem.parent = elems[elem.t_parent] if elem.t_parent else None
        controllers = [
            mocker.Mock(root='a', elem=None),
            mocker.Mock(root='c', elem='b'),
        ]
        parent = None
        for cont in controllers:
            cont._micropath_parent = parent
            cont._micropath_elem = (
                elems[cont.elem] if cont.elem else None
            )
            root = mocker.Mock(spec=elements.Root)
            root.parent = None
            elems[cont.root].parent = root
            parent = cont
        meth = mocker.Mock(_micropath_elem=elems['d'])
        mock_get_method_self = mocker.patch.object(
            request.six, 'get_method_self',
            return_value=controllers[-1],
        )
        mock_isclass = mocker.patch.object(
            request.inspect, 'isclass',
            return_value=False,
        )
        obj = request.Request.blank(
            '/is/a/test',
            base_url='http://example.com/this',
        )

        with pytest.raises(ValueError):
            obj.url_for(meth, b=1)
        mock_get_method_self.assert_called_once_with(meth)
        mock_isclass.assert_called_once_with(controllers[-1])

    def test_injector_cached(self, mocker):
        mock_Injector = mocker.patch.object(request.injector, 'Injector')
        obj = request.Request.blank('/')
        obj.environ['micropath.injector'] = 'cached'

        assert obj.injector == 'cached'
        assert obj.environ['micropath.injector'] == 'cached'
        mock_Injector.assert_not_called()

    def test_injector_uncached(self, mocker):
        mock_Injector = mocker.patch.object(request.injector, 'Injector')
        obj = request.Request.blank('/')

        assert obj.injector == mock_Injector.return_value
        assert obj.environ['micropath.injector'] == mock_Injector.return_value
        mock_Injector.assert_called_once_with()

    def test_base_path_get(self):
        obj = request.Request.blank('/')
        obj.environ['micropath.base_path'] = '/base/path'

        assert obj.base_path == '/base/path'



# Copyright (C) 2018 by Kevin L. Mitchell <klmitch@mit.edu>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you
# may not use this file except in compliance with the License. You may
# obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied. See the License for the specific language governing
# permissions and limitations under the License.

import inspect

import six
import webob

from micropath import elements
from micropath import injector


class Request(webob.Request):
    """
    A subclass of ``webob.Request`` containing additional support used
    by the ``micropath`` framework.  In particular, the ``injector``
    attribute contains the dependency injector used by ``micropath``
    to invoke handler methods, and the ``base_path`` attribute is the
    value of the ``SCRIPT_NAME`` WSGI environment variable at the time
    the ``Request`` was constructed.  (This latter attribute may be
    used to construct absolute paths to other ``micropath`` handlers.)
    In addition, the ``url_for()`` method is capable of constructing a
    URL for any given controller method.
    """

    def __init__(self, *args, **kwargs):
        """
        Initialize a ``Request`` instance.  Parameters are passed to the
        underlying ``webob.Request``.  This method also adds the
        ``micropath.base_path`` key to the WSGI environment (use the
        ``base_path`` attribute to access it).
        """

        super(Request, self).__init__(*args, **kwargs)

        # Save the base path for later URL computation
        self.environ['micropath.base_path'] = self.script_name

    def url_for(self, *args, **kwargs):
        """
        Construct the absolute URL for a given handler method.  The
        handler method should be passed as the only positional
        parameter, and keyword parameters should be passed to provide
        values for the bindings (which will be passed to the
        ``micropath.elements.Binding.format()`` method to be converted
        into URL path elements).

        :returns: An absolute URL corresponding to the handler method.
                  Note that it is not guaranteed that that handler
                  method will be called by accessing the URL; that
                  depends on the HTTP methods routed to the handler
                  method.
        :rtype: ``str``
        """

        # Sanity-check the arguments
        if len(args) != 1:
            raise TypeError(
                'url_for() requires exactly 1 positional argument; %d given' %
                len(args),
            )
        elif (not callable(args[0]) or
              getattr(args[0], '_micropath_elem', None) is None):
            raise ValueError('unable to construct URL for %r' % args[0])

        # Get the controller and element
        controller = six.get_method_self(args[0])
        elem = args[0]._micropath_elem

        # Make sure it is a controller instance and not a class
        if inspect.isclass(controller):
            raise ValueError('unable to construct URL for class method')

        # Begin walking up the controller/element tree
        path_elems = []
        while controller:
            while elem:
                if isinstance(elem, elements.Path):
                    # Just add the element's ident
                    path_elems.append(elem.ident)
                elif isinstance(elem, elements.Binding):
                    # Make sure we have a value
                    if elem.ident not in kwargs:
                        raise ValueError(
                            'missing value for binding "%s"' % elem.ident,
                        )

                    # Format the value
                    path_elems.append(elem.format(
                        controller, kwargs[elem.ident],
                    ))
                else:
                    assert isinstance(elem, elements.Root)

                # Walk up to the element's parent
                elem = elem.parent

            # Walk up the controller chain
            elem = getattr(controller, '_micropath_elem', None)
            controller = getattr(controller, '_micropath_parent', None)

        # Construct the path
        return '%s%s/%s' % (
            self.host_url, self.base_path, '/'.join(reversed(path_elems)),
        )

    @property
    def injector(self):
        """
        Retrieve the dependency injector from the WSGI environment.
        """

        if 'micropath.injector' not in self.environ:
            self.environ['micropath.injector'] = injector.Injector()

        return self.environ['micropath.injector']

    @property
    def base_path(self):
        """
        Retrieve the base path.  This may be used to construct full URLs.
        The base path is the value of the ``SCRIPT_NAME`` WSGI
        environment variable at the time the ``Request`` instance is
        constructed.
        """

        return self.environ.get('micropath.base_path')



# Copyright (C) 2018 by Kevin L. Mitchell <klmitch@mit.edu>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you
# may not use this file except in compliance with the License. You may
# obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied. See the License for the specific language governing
# permissions and limitations under the License.

import pytest

from micropath import elements


class ElementForTest(elements.Element):
    def set_ident(self, ident):
        super(ElementForTest, self).set_ident(ident)


class OtherElement(elements.Element):
    def set_ident(self, ident):
        pass


class TestElement(object):
    def test_init_base(self):
        result = ElementForTest('ident')

        assert result.ident == 'ident'
        assert result.parent is None
        assert result.paths == {}
        assert result.bindings is None
        assert result.methods == {}
        assert result.delegation is None

    def test_init_alt(self):
        result = ElementForTest('ident', 'parent')

        assert result.ident == 'ident'
        assert result.parent == 'parent'
        assert result.paths == {}
        assert result.bindings is None
        assert result.methods == {}
        assert result.delegation is None

    def test_set_ident_base(self):
        obj = ElementForTest(None)

        obj.set_ident('ident')

        assert obj.ident == 'ident'

    def test_set_ident_set(self):
        obj = ElementForTest('ident')

        with pytest.raises(ValueError):
            obj.set_ident('spam')
        assert obj.ident == 'ident'

    @staticmethod
    def sub_sel(subs, *elems):
        return {elem: subs[elem] for elem in elems}

    def test_path_base(self, mocker):
        mock_Path = mocker.patch.object(
            elements, 'Path',
            return_value=mocker.Mock(ident=None),
        )
        obj = ElementForTest('ident')

        result = obj.path()

        assert result == mock_Path.return_value
        mock_Path.assert_called_once_with(None, parent=obj)
        assert obj.paths == {}

    def test_path_with_ident(self, mocker):
        mock_Path = mocker.patch.object(
            elements, 'Path',
            return_value=mocker.Mock(ident='spam'),
        )
        obj = ElementForTest('ident')

        result = obj.path('spam')

        assert result == mock_Path.return_value
        mock_Path.assert_called_once_with('spam', parent=obj)
        assert obj.paths == {'spam': result}

    def test_path_conflict(self, mocker):
        mock_Path = mocker.patch.object(
            elements, 'Path',
            return_value=mocker.Mock(ident='spam'),
        )
        obj = ElementForTest('ident')
        obj.paths['spam'] = 'conflict'

        with pytest.raises(ValueError):
            obj.path('spam')
        mock_Path.assert_called_once_with('spam', parent=obj)
        assert obj.paths == {'spam': 'conflict'}

    def test_binding_base(self, mocker):
        mock_Binding = mocker.patch.object(
            elements, 'Binding',
            return_value=mocker.Mock(ident=None),
        )
        obj = ElementForTest('ident')

        result = obj.bind()

        assert result == mock_Binding.return_value
        mock_Binding.assert_called_once_with(
            None,
            parent=obj,
        )
        assert obj.bindings is None

    def test_binding_with_ident(self, mocker):
        mock_Binding = mocker.patch.object(
            elements, 'Binding',
            return_value=mocker.Mock(ident='spam'),
        )
        obj = ElementForTest('ident')

        result = obj.bind('spam')

        assert result == mock_Binding.return_value
        mock_Binding.assert_called_once_with(
            'spam',
            parent=obj,
        )
        assert obj.bindings == result

    def test_binding_conflict(self, mocker):
        mock_Binding = mocker.patch.object(
            elements, 'Binding',
            return_value=mocker.Mock(ident='spam'),
        )
        obj = ElementForTest('ident')
        obj.bindings = 'conflict'

        with pytest.raises(ValueError):
            obj.bind('spam')
        mock_Binding.assert_called_once_with(
            'spam',
            parent=obj,
        )
        assert obj.bindings == 'conflict'

    def test_route_func(self, mocker):
        mock_Method = mocker.patch.object(
            elements, 'Method',
            return_value=mocker.Mock(ident=None),
        )
        mock_from_func = mocker.patch.object(
            elements.injector.WantSignature, 'from_func',
        )
        obj = ElementForTest('ident')
        func = mocker.Mock(_micropath_handler=False)

        result = obj.route(func)

        assert result == func
        mock_Method.assert_called_once_with(None, func, parent=obj)
        mock_from_func.assert_called_once_with(func)
        assert obj.methods == {None: mock_Method.return_value}
        assert func._micropath_handler is True
        assert func._micropath_elem is obj

    def test_route_no_methods(self, mocker):
        mock_Method = mocker.patch.object(
            elements, 'Method',
            return_value=mocker.Mock(ident=None),
        )
        mock_from_func = mocker.patch.object(
            elements.injector.WantSignature, 'from_func',
        )
        obj = ElementForTest('ident')
        func = mocker.Mock(_micropath_handler=False)

        decorator = obj.route()

        assert callable(decorator)
        assert decorator != func
        mock_Method.assert_not_called()
        mock_from_func.assert_not_called()
        assert obj.methods == {}

        result = decorator(func)

        assert result == func
        mock_Method.assert_called_once_with(None, func, parent=obj)
        mock_from_func.assert_called_once_with(func)
        assert obj.methods == {None: mock_Method.return_value}
        assert func._micropath_handler is True
        assert func._micropath_elem is obj

    def test_route_with_methods(self, mocker):
        methods = {
            'get': mocker.Mock(ident='get'),
            'put': mocker.Mock(ident='put'),
        }
        mock_Method = mocker.patch.object(
            elements, 'Method',
            side_effect=lambda x, f, parent: methods[x],
        )
        mock_from_func = mocker.patch.object(
            elements.injector.WantSignature, 'from_func',
        )
        obj = ElementForTest('ident')
        func = mocker.Mock(_micropath_handler=False)

        decorator = obj.route('get', 'put')

        assert callable(decorator)
        assert decorator != func
        mock_Method.assert_not_called()
        mock_from_func.assert_not_called()
        assert obj.methods == {}

        result = decorator(func)

        assert result == func
        mock_Method.assert_has_calls([
            mocker.call('get', func, parent=obj),
            mocker.call('put', func, parent=obj),
        ])
        assert mock_Method.call_count == 2
        mock_from_func.assert_called_once_with(func)
        assert obj.methods == methods
        assert func._micropath_handler is True
        assert func._micropath_elem is obj

    def test_route_with_methods_internal_duplicate(self, mocker):
        methods = {
            'get': mocker.Mock(ident='get'),
            'put': mocker.Mock(ident='put'),
        }
        mock_Method = mocker.patch.object(
            elements, 'Method',
            side_effect=lambda x, f, parent: methods[x],
        )
        mock_from_func = mocker.patch.object(
            elements.injector.WantSignature, 'from_func',
        )
        obj = ElementForTest('ident')
        func = mocker.Mock(_micropath_handler=False)

        decorator = obj.route('get', 'put', 'get')

        assert callable(decorator)
        assert decorator != func
        mock_Method.assert_not_called()
        mock_from_func.assert_not_called()
        assert obj.methods == {}

        result = decorator(func)

        assert result == func
        mock_Method.assert_has_calls([
            mocker.call('get', func, parent=obj),
            mocker.call('put', func, parent=obj),
        ])
        assert mock_Method.call_count == 2
        mock_from_func.assert_called_once_with(func)
        assert obj.methods == methods
        assert func._micropath_handler is True
        assert func._micropath_elem is obj

    def test_route_with_methods_external_duplicate(self, mocker):
        methods = {
            'get': mocker.Mock(ident='get'),
            'put': mocker.Mock(ident='put'),
        }
        mock_Method = mocker.patch.object(
            elements, 'Method',
            side_effect=lambda x, f, parent: methods[x],
        )
        mock_from_func = mocker.patch.object(
            elements.injector.WantSignature, 'from_func',
        )
        obj = ElementForTest('ident')
        obj.methods['get'] = 'conflict'

        with pytest.raises(ValueError):
            obj.route('get', 'put')
        mock_Method.assert_not_called()
        mock_from_func.assert_not_called()
        assert obj.methods == {'get': 'conflict'}

    def test_mount_base(self, mocker):
        mock_init = mocker.patch.object(
            elements.Delegation, '__init__',
            return_value=None,
        )
        mock_Method = mocker.patch.object(
            elements, 'Method',
            return_value=mocker.Mock(ident=None),
        )
        obj = ElementForTest('ident')

        result = obj.mount('delegation')

        assert isinstance(result, elements.Delegation)
        assert result.element == obj
        assert obj.methods == {}
        assert obj.delegation == result
        mock_init.assert_called_once_with('delegation', {})
        mock_Method.assert_not_called()

    def test_mount_with_methods(self, mocker):
        methods = {
            'get': mocker.Mock(ident='get', delegation=None),
            'put': mocker.Mock(ident='put', delegation=None),
        }
        mock_init = mocker.patch.object(
            elements.Delegation, '__init__',
            return_value=None,
        )
        mock_Method = mocker.patch.object(
            elements, 'Method',
            side_effect=lambda x, f, parent: methods[x],
        )
        obj = ElementForTest('ident')

        result = obj.mount('delegation', 'get', 'put', a=1, b=2)

        assert isinstance(result, elements.Delegation)
        assert result.element == obj
        assert obj.methods == methods
        for meth in methods.values():
            assert meth.delegation == result
        assert obj.delegation is None
        mock_init.assert_called_once_with('delegation', {'a': 1, 'b': 2})
        mock_Method.assert_has_calls([
            mocker.call('get', None, parent=obj),
            mocker.call('put', None, parent=obj),
        ])
        assert mock_Method.call_count == 2

    def test_mount_with_methods_internal_duplication(self, mocker):
        methods = {
            'get': mocker.Mock(ident='get', delegation=None),
            'put': mocker.Mock(ident='put', delegation=None),
        }
        mock_init = mocker.patch.object(
            elements.Delegation, '__init__',
            return_value=None,
        )
        mock_Method = mocker.patch.object(
            elements, 'Method',
            side_effect=lambda x, f, parent: methods[x],
        )
        obj = ElementForTest('ident')

        result = obj.mount('delegation', 'get', 'put', 'get', a=1, b=2)

        assert isinstance(result, elements.Delegation)
        assert result.element == obj
        assert obj.methods == methods
        for meth in methods.values():
            assert meth.delegation == result
        assert obj.delegation is None
        mock_init.assert_called_once_with('delegation', {'a': 1, 'b': 2})
        mock_Method.assert_has_calls([
            mocker.call('get', None, parent=obj),
            mocker.call('put', None, parent=obj),
        ])
        assert mock_Method.call_count == 2

    def test_mount_with_methods_external_duplication(self, mocker):
        methods = {
            'get': mocker.Mock(ident='get', delegation=None),
            'put': mocker.Mock(ident='put', delegation=None),
        }
        mock_init = mocker.patch.object(
            elements.Delegation, '__init__',
            return_value=None,
        )
        mock_Method = mocker.patch.object(
            elements, 'Method',
            side_effect=lambda x, f, parent: methods[x],
        )
        obj = ElementForTest('ident')
        obj.methods['get'] = 'conflict'

        with pytest.raises(ValueError):
            obj.mount('delegation', 'get', 'put', a=1, b=2)
        assert obj.methods == {'get': 'conflict'}
        for meth in methods.values():
            assert meth.delegation is None
        assert obj.delegation is None
        mock_init.assert_called_once_with('delegation', {'a': 1, 'b': 2})
        mock_Method.assert_not_called()

    def test_mount_delegation(self, mocker):
        delegation = elements.Delegation('delegation', {})
        mock_init = mocker.patch.object(
            elements.Delegation, '__init__',
            return_value=None,
        )
        mock_Method = mocker.patch.object(
            elements, 'Method',
            return_value=mocker.Mock(ident=None),
        )
        obj = ElementForTest('ident')

        result = obj.mount(delegation)

        assert result == delegation
        assert result.element == obj
        assert obj.methods == {}
        assert obj.delegation == delegation
        mock_init.assert_not_called()
        mock_Method.assert_not_called()

    def test_mount_delegation_set(self, mocker):
        mock_init = mocker.patch.object(
            elements.Delegation, '__init__',
            return_value=None,
        )
        mock_Method = mocker.patch.object(
            elements, 'Method',
            return_value=mocker.Mock(ident=None),
        )
        obj = ElementForTest('ident')
        obj.delegation = 'spam'

        with pytest.raises(ValueError):
            obj.mount('delegation')
        assert obj.methods == {}
        assert obj.delegation == 'spam'
        mock_init.assert_not_called()
        mock_Method.assert_not_called()


class TestRoot(object):
    def test_init(self, mocker):
        mock_init = mocker.patch.object(
            elements.Element, '__init__',
            return_value=None,
        )

        result = elements.Root()

        assert isinstance(result, elements.Root)
        mock_init.assert_called_once_with(None)

    def test_set_ident(self):
        obj = elements.Root()

        with pytest.raises(ValueError):
            obj.set_ident('ident')

    def test_add_elem_path(self, mocker):
        elem = mocker.Mock(spec=elements.Path, ident='spam')
        elem.parent = None
        obj = elements.Root()

        obj.add_elem(elem)

        assert obj.paths == {'spam': elem}
        assert obj.bindings is None
        assert obj.methods == {}
        assert elem.ident == 'spam'
        assert elem.parent is obj
        elem.set_ident.assert_not_called()

    def test_add_elem_path_conflict(self, mocker):
        elem = mocker.Mock(spec=elements.Path, ident='spam')
        elem.parent = None
        obj = elements.Root()
        obj.paths['spam'] = 'conflict'

        with pytest.raises(ValueError):
            obj.add_elem(elem)
        assert obj.paths == {'spam': 'conflict'}
        assert obj.bindings is None
        assert obj.methods == {}
        assert elem.ident == 'spam'
        assert elem.parent is None
        elem.set_ident.assert_not_called()

    def test_add_elem_binding(self, mocker):
        elem = mocker.Mock(spec=elements.Binding, ident='spam')
        elem.parent = None
        obj = elements.Root()

        obj.add_elem(elem)

        assert obj.paths == {}
        assert obj.bindings == elem
        assert obj.methods == {}
        assert elem.ident == 'spam'
        assert elem.parent is obj
        elem.set_ident.assert_not_called()

    def test_add_elem_binding_conflict(self, mocker):
        elem = mocker.Mock(spec=elements.Binding, ident='spam')
        elem.parent = None
        obj = elements.Root()
        obj.bindings = 'conflict'

        with pytest.raises(ValueError):
            obj.add_elem(elem)
        assert obj.paths == {}
        assert obj.bindings == 'conflict'
        assert obj.methods == {}
        assert elem.ident == 'spam'
        assert elem.parent is None
        elem.set_ident.assert_not_called()

    def test_add_elem_method(self, mocker):
        elem = mocker.Mock(spec=elements.Method, ident='spam')
        elem.parent = None
        obj = elements.Root()

        obj.add_elem(elem)

        assert obj.paths == {}
        assert obj.bindings is None
        assert obj.methods == {'spam': elem}
        assert elem.ident == 'spam'
        assert elem.parent is obj
        elem.set_ident.assert_not_called()

    def test_add_elem_method_conflict(self, mocker):
        elem = mocker.Mock(spec=elements.Method, ident='spam')
        elem.parent = None
        obj = elements.Root()
        obj.methods['spam'] = 'conflict'

        with pytest.raises(ValueError):
            obj.add_elem(elem)
        assert obj.paths == {}
        assert obj.bindings is None
        assert obj.methods == {'spam': 'conflict'}
        assert elem.ident == 'spam'
        assert elem.parent is None
        elem.set_ident.assert_not_called()

    def test_add_elem_method_all(self, mocker):
        elem = mocker.Mock(spec=elements.Method, ident=None)
        elem.parent = None
        obj = elements.Root()

        obj.add_elem(elem)

        assert obj.paths == {}
        assert obj.bindings is None
        assert obj.methods == {None: elem}
        assert elem.ident is None
        assert elem.parent is obj
        elem.set_ident.assert_not_called()

    def test_add_elem_method_all_conflict(self, mocker):
        elem = mocker.Mock(spec=elements.Method, ident=None)
        elem.parent = None
        obj = elements.Root()
        obj.methods[None] = 'conflict'

        with pytest.raises(ValueError):
            obj.add_elem(elem)
        assert obj.paths == {}
        assert obj.bindings is None
        assert obj.methods == {None: 'conflict'}
        assert elem.ident is None
        assert elem.parent is None
        elem.set_ident.assert_not_called()

    def test_add_elem_other(self, mocker):
        elem = mocker.Mock(ident='spam')
        elem.parent = None
        obj = elements.Root()

        with pytest.raises(ValueError):
            obj.add_elem(elem)
        assert obj.paths == {}
        assert obj.bindings is None
        assert obj.methods == {}
        assert elem.ident == 'spam'
        assert elem.parent is None
        elem.set_ident.assert_not_called()

    def test_add_elem_self(self, mocker):
        obj = elements.Root()

        obj.add_elem(obj)

        assert obj.paths == {}
        assert obj.bindings is None
        assert obj.methods == {}

    def test_add_elem_root(self, mocker):
        elem = mocker.Mock(spec=elements.Root, ident=None)
        elem.parent = None
        obj = elements.Root()

        with pytest.raises(ValueError):
            obj.add_elem(elem)
        assert obj.paths == {}
        assert obj.bindings is None
        assert obj.methods == {}
        assert elem.ident is None
        assert elem.parent is None
        elem.set_ident.assert_not_called()

    def test_add_elem_path_no_ident(self, mocker):
        elem = mocker.Mock(spec=elements.Path, ident=None)
        elem.parent = None
        obj = elements.Root()

        obj.add_elem(elem)

        assert obj.paths == {}
        assert obj.bindings is None
        assert obj.methods == {}
        assert elem.ident is None
        assert elem.parent is obj
        elem.set_ident.assert_not_called()

    def test_add_elem_binding_no_ident(self, mocker):
        elem = mocker.Mock(spec=elements.Binding, ident=None)
        elem.parent = None
        obj = elements.Root()

        obj.add_elem(elem)

        assert obj.paths == {}
        assert obj.bindings is None
        assert obj.methods == {}
        assert elem.ident is None
        assert elem.parent is obj
        elem.set_ident.assert_not_called()

    def test_add_elem_set_ident(self, mocker):
        elem = mocker.Mock(spec=elements.Path, ident=None)
        elem.parent = None
        obj = elements.Root()

        obj.add_elem(elem, 'spam')

        assert obj.paths == {}
        assert obj.bindings is None
        assert obj.methods == {}
        assert elem.ident is None
        assert elem.parent is obj
        elem.set_ident.assert_called_once_with('spam')

    def test_add_elem_parents(self, mocker):
        elem = mocker.Mock(spec=elements.Path, ident='spam')
        elem.parent = None
        descendant = mocker.Mock(spec=elements.Path, ident=None)
        descendant.parent = elem
        obj = elements.Root()

        obj.add_elem(descendant, 'descendant')

        assert obj.paths == {'spam': elem}
        assert obj.bindings is None
        assert obj.methods == {}
        assert elem.ident == 'spam'
        assert elem.parent is obj
        elem.set_ident.assert_not_called()
        descendant.set_ident.assert_called_once_with('descendant')


class TestPath(object):
    def test_set_ident_no_parent(self, mocker):
        mock_set_ident = mocker.patch.object(elements.Element, 'set_ident')
        obj = elements.Path(None)

        obj.set_ident('ident')

        mock_set_ident.assert_called_once_with('ident')

    def test_set_ident_with_parent(self, mocker):
        mock_set_ident = mocker.patch.object(elements.Element, 'set_ident')
        obj = elements.Path(None)
        obj.parent = mocker.Mock(paths={})

        obj.set_ident('ident')

        assert obj.parent.paths == {None: obj}
        mock_set_ident.assert_called_once_with('ident')

    def test_set_ident_conflict(self, mocker):
        mock_set_ident = mocker.patch.object(elements.Element, 'set_ident')
        obj = elements.Path(None)
        obj.parent = mocker.Mock(paths={None: 'conflict'})

        with pytest.raises(ValueError):
            obj.set_ident('ident')
        assert obj.parent.paths == {None: 'conflict'}
        mock_set_ident.assert_called_once_with('ident')


class TestBinding(object):
    def test_init_base(self, mocker):
        mock_init = mocker.patch.object(
            elements.Element, '__init__',
            return_value=None,
        )

        result = elements.Binding('ident')

        assert result._validator is None
        assert result._formatter is None
        mock_init.assert_called_once_with('ident', None)

    def test_init_alt(self, mocker):
        mock_init = mocker.patch.object(
            elements.Element, '__init__',
            return_value=None,
        )

        result = elements.Binding(
            'ident', 'parent',
        )

        assert result._validator is None
        assert result._formatter is None
        mock_init.assert_called_once_with('ident', 'parent')

    def test_set_ident_no_parent(self, mocker):
        mock_set_ident = mocker.patch.object(elements.Element, 'set_ident')
        obj = elements.Binding(None)

        obj.set_ident('ident')

        mock_set_ident.assert_called_once_with('ident')

    def test_set_ident_with_parent(self, mocker):
        mock_set_ident = mocker.patch.object(elements.Element, 'set_ident')
        obj = elements.Binding(None)
        obj.parent = mocker.Mock(bindings=None)

        obj.set_ident('ident')

        assert obj.parent.bindings == obj
        mock_set_ident.assert_called_once_with('ident')

    def test_set_ident_conflict(self, mocker):
        mock_set_ident = mocker.patch.object(elements.Element, 'set_ident')
        obj = elements.Binding(None)
        obj.parent = mocker.Mock(bindings='conflict')

        with pytest.raises(ValueError):
            obj.set_ident('ident')
        assert obj.parent.bindings == 'conflict'
        mock_set_ident.assert_called_once_with('ident')

    def test_validator_base(self, mocker):
        mock_from_func = mocker.patch.object(
            elements.injector.WantSignature, 'from_func',
        )
        obj = elements.Binding('ident')

        result = obj.validator('func')

        assert result == 'func'
        assert obj._validator == 'func'
        mock_from_func.assert_called_once_with('func')

    def test_validator_already_set(self, mocker):
        mock_from_func = mocker.patch.object(
            elements.injector.WantSignature, 'from_func',
        )
        obj = elements.Binding('ident')
        obj._validator = 'spam'

        with pytest.raises(ValueError):
            obj.validator('func')
        assert obj._validator == 'spam'
        mock_from_func.assert_not_called()

    def test_validate_unset(self, mocker):
        inj = mocker.Mock()
        obj = elements.Binding('ident')

        result = obj.validate('controller', inj, 'value')

        assert result == 'value'
        inj.assert_not_called()

    def test_validate_set(self, mocker):
        inj = mocker.Mock()
        obj = elements.Binding('ident')
        obj._validator = 'validator'

        result = obj.validate('controller', inj, 'value')

        assert result == inj.return_value
        inj.assert_called_once_with('validator', 'controller', value='value')

    def test_formatter_base(self):
        obj = elements.Binding('ident')

        result = obj.formatter('func')

        assert result == 'func'
        assert obj._formatter == 'func'

    def test_formatter_already_set(self):
        obj = elements.Binding('ident')
        obj._formatter = 'spam'

        with pytest.raises(ValueError):
            obj.formatter('func')
        assert obj._formatter == 'spam'

    def test_format_unset(self):
        obj = elements.Binding('ident')

        result = obj.format('controller', 1234)

        assert result == '1234'

    def test_format_set(self, mocker):
        obj = elements.Binding('ident')
        obj._formatter = mocker.Mock(return_value='string')

        result = obj.format('controller', 1234)

        assert result == 'string'
        obj._formatter.assert_called_once_with('controller', 1234)


class TestMethod(object):
    def test_init_base(self, mocker):
        mock_init = mocker.patch.object(
            elements.Element, '__init__',
            return_value=None,
        )

        result = elements.Method('get', 'func')

        assert result.func == 'func'
        mock_init.assert_called_once_with('GET', None)

    def test_init_alt(self, mocker):
        mock_init = mocker.patch.object(
            elements.Element, '__init__',
            return_value=None,
        )

        result = elements.Method(None, 'func', 'parent')

        assert result.func == 'func'
        mock_init.assert_called_once_with(None, 'parent')

    def test_set_ident(self):
        obj = elements.Method(None, 'func')

        with pytest.raises(ValueError):
            obj.set_ident('ident')

    def test_path_base(self):
        obj = elements.Method(None, 'func')

        with pytest.raises(ValueError):
            obj.path()

    def test_path_alt(self):
        obj = elements.Method(None, 'func')

        with pytest.raises(ValueError):
            obj.path('ident')

    def test_bind_base(self):
        obj = elements.Method(None, 'func')

        with pytest.raises(ValueError):
            obj.bind()

    def test_bind_alt(self):
        obj = elements.Method(None, 'func')

        with pytest.raises(ValueError):
            obj.bind('ident')

    def test_route_base(self):
        obj = elements.Method(None, 'func')

        with pytest.raises(ValueError):
            obj.route()

    def test_route_alt(self):
        obj = elements.Method(None, 'func')

        with pytest.raises(ValueError):
            obj.route('get', 'put')

    def test_mount(self, mocker):
        mock_mount = mocker.patch.object(elements.Element, 'mount')
        obj = elements.Method(None, 'func')

        result = obj.mount('delegation')

        assert result == mock_mount.return_value
        mock_mount.assert_called_once_with('delegation')


class TestDelegation(object):
    def test_init(self):
        result = elements.Delegation('controller', 'kwargs')

        assert result.controller == 'controller'
        assert result.kwargs == 'kwargs'
        assert result.element is None
        assert result._cache == {}

    def test_dunder_get_class(self, mocker):
        mock_get = mocker.patch.object(elements.Delegation, 'get')
        obj = elements.Delegation('controller', {})

        result = obj.__get__(None, 'class')

        assert result is obj
        mock_get.assert_not_called()

    def test_dunder_get_object(self, mocker):
        mock_get = mocker.patch.object(elements.Delegation, 'get')
        obj = elements.Delegation('controller', {})

        result = obj.__get__('object', 'class')

        assert result == mock_get.return_value
        mock_get.assert_called_once_with('object')

    def test_set(self):
        target = object()
        obj = elements.Delegation('controller', {})

        obj.__set__(target, 'value')

        assert obj._cache == {id(target): 'value'}

    def test_delete_exists(self):
        target = object()
        obj = elements.Delegation('controller', {})
        obj._cache = {id(target): 'value'}

        obj.__delete__(target)

        assert obj._cache == {}

    def test_delete_missing(self):
        target = object()
        obj = elements.Delegation('controller', {})

        obj.__delete__(target)

        assert obj._cache == {}

    def test_get_cached(self, mocker):
        mock_construct = mocker.patch.object(elements.Delegation, 'construct')
        target = object()
        obj = elements.Delegation('controller', {})
        obj._cache = {id(target): 'value'}

        result = obj.get(target)

        assert result == 'value'
        assert obj._cache == {id(target): 'value'}
        mock_construct.assert_not_called()

    def test_get_uncached(self, mocker):
        mock_construct = mocker.patch.object(elements.Delegation, 'construct')
        target = object()
        obj = elements.Delegation('controller', {})
        obj.element = 'element'

        result = obj.get(target)

        assert result == mock_construct.return_value
        assert obj._cache == {id(target): mock_construct.return_value}
        assert mock_construct.return_value._micropath_parent is target
        assert mock_construct.return_value._micropath_elem == 'element'
        mock_construct.assert_called_once_with(target)

    def test_construct(self, mocker):
        target = mocker.Mock()
        obj = elements.Delegation('controller', 'kwargs')

        result = obj.construct(target)

        assert result == target.micropath_construct.return_value
        target.micropath_construct.assert_called_once_with(
            'controller', 'kwargs',
        )


class TestPathFunc(object):
    def test_base(self, mocker):
        mock_Path = mocker.patch.object(elements, 'Path')

        result = elements.path()

        assert result == mock_Path.return_value
        mock_Path.assert_called_once_with(None)

    def test_alt(self, mocker):
        mock_Path = mocker.patch.object(elements, 'Path')

        result = elements.path('ident')

        assert result == mock_Path.return_value
        mock_Path.assert_called_once_with('ident')


class TestBind(object):
    def test_base(self, mocker):
        mock_Binding = mocker.patch.object(elements, 'Binding')

        result = elements.bind()

        assert result == mock_Binding.return_value
        mock_Binding.assert_called_once_with(None)

    def test_alt(self, mocker):
        mock_Binding = mocker.patch.object(elements, 'Binding')

        result = elements.bind('ident')

        assert result == mock_Binding.return_value
        mock_Binding.assert_called_once_with(
            'ident',
        )


class TestRoute(object):
    def test_func(self, mocker):
        mock_Method = mocker.patch.object(
            elements, 'Method',
            return_value=mocker.Mock(ident=None),
        )
        mock_from_func = mocker.patch.object(
            elements.injector.WantSignature, 'from_func',
        )
        func = mocker.Mock(_micropath_methods=None, _micropath_handler=False)

        result = elements.route(func)

        assert result == func
        mock_Method.assert_called_once_with(None, func)
        mock_from_func.assert_called_once_with(func)
        assert func._micropath_methods == [mock_Method.return_value]
        assert func._micropath_handler is True

    def test_no_methods(self, mocker):
        mock_Method = mocker.patch.object(
            elements, 'Method',
            return_value=mocker.Mock(ident=None),
        )
        mock_from_func = mocker.patch.object(
            elements.injector.WantSignature, 'from_func',
        )
        func = mocker.Mock(_micropath_methods=None, _micropath_handler=False)

        decorator = elements.route()

        assert callable(decorator)
        assert decorator != func
        mock_Method.assert_not_called()
        mock_from_func.assert_not_called()
        assert func._micropath_methods is None
        assert func._micropath_handler is False

        result = decorator(func)

        assert result == func
        mock_Method.assert_called_once_with(None, func)
        mock_from_func.assert_called_once_with(func)
        assert func._micropath_methods == [mock_Method.return_value]
        assert func._micropath_handler is True

    def test_with_methods(self, mocker):
        methods = {
            'get': mocker.Mock(ident='get'),
            'put': mocker.Mock(ident='put'),
        }
        mock_Method = mocker.patch.object(
            elements, 'Method',
            side_effect=lambda x, f: methods[x],
        )
        mock_from_func = mocker.patch.object(
            elements.injector.WantSignature, 'from_func',
        )
        func = mocker.Mock(_micropath_methods=None, _micropath_handler=False)

        decorator = elements.route('get', 'put')

        assert callable(decorator)
        assert decorator != func
        mock_Method.assert_not_called()
        mock_from_func.assert_not_called()
        assert func._micropath_methods is None
        assert func._micropath_handler is False

        result = decorator(func)

        assert result == func
        mock_Method.assert_has_calls([
            mocker.call('get', func),
            mocker.call('put', func),
        ])
        assert mock_Method.call_count == 2
        mock_from_func.assert_called_once_with(func)
        assert func._micropath_methods == [methods[x] for x in ('get', 'put')]
        assert func._micropath_handler is True

    def test_with_methods_internal_duplicate(self, mocker):
        methods = {
            'get': mocker.Mock(ident='get'),
            'put': mocker.Mock(ident='put'),
        }
        mock_Method = mocker.patch.object(
            elements, 'Method',
            side_effect=lambda x, f: methods[x],
        )
        mock_from_func = mocker.patch.object(
            elements.injector.WantSignature, 'from_func',
        )
        func = mocker.Mock(_micropath_methods=None, _micropath_handler=False)

        decorator = elements.route('get', 'put', 'get')

        assert callable(decorator)
        assert decorator != func
        mock_Method.assert_not_called()
        mock_from_func.assert_not_called()
        assert func._micropath_methods is None
        assert func._micropath_handler is False

        result = decorator(func)

        assert result == func
        mock_Method.assert_has_calls([
            mocker.call('get', func),
            mocker.call('put', func),
        ])
        assert mock_Method.call_count == 2
        mock_from_func.assert_called_once_with(func)
        assert func._micropath_methods == [methods[x] for x in ('get', 'put')]
        assert func._micropath_handler is True


class TestMount(object):
    def test_base(self, mocker):
        mock_init = mocker.patch.object(
            elements.Delegation, '__init__',
            return_value=None,
        )
        mock_Method = mocker.patch.object(
            elements, 'Method',
            return_value=mocker.Mock(ident=None),
        )

        result = elements.mount('delegation')

        assert isinstance(result, elements.Delegation)
        assert not hasattr(result, '_micropath_methods')
        mock_init.assert_called_once_with('delegation', {})
        mock_Method.assert_not_called()

    def test_with_methods(self, mocker):
        methods = {
            'get': mocker.Mock(ident='get', delegation=None),
            'put': mocker.Mock(ident='put', delegation=None),
        }
        mock_init = mocker.patch.object(
            elements.Delegation, '__init__',
            return_value=None,
        )
        mock_Method = mocker.patch.object(
            elements, 'Method',
            side_effect=lambda x, f: methods[x],
        )

        result = elements.mount('delegation', 'get', 'put', a=1, b=2)

        assert isinstance(result, elements.Delegation)
        assert result._micropath_methods == [methods['get'], methods['put']]
        mock_init.assert_called_once_with('delegation', {'a': 1, 'b': 2})
        mock_Method.assert_has_calls([
            mocker.call('get', None),
            mocker.call('put', None),
        ])

    def test_with_methods_internal_duplication(self, mocker):
        methods = {
            'get': mocker.Mock(ident='get', delegation=None),
            'put': mocker.Mock(ident='put', delegation=None),
        }
        mock_init = mocker.patch.object(
            elements.Delegation, '__init__',
            return_value=None,
        )
        mock_Method = mocker.patch.object(
            elements, 'Method',
            side_effect=lambda x, f: methods[x],
        )

        result = elements.mount('delegation', 'get', 'put', 'get', a=1, b=2)

        assert isinstance(result, elements.Delegation)
        assert result._micropath_methods == [methods['get'], methods['put']]
        mock_init.assert_called_once_with('delegation', {'a': 1, 'b': 2})
        mock_Method.assert_has_calls([
            mocker.call('get', None),
            mocker.call('put', None),
        ])

    def test_delegation(self, mocker):
        delegation = elements.Delegation('delegation', {})
        mock_init = mocker.patch.object(
            elements.Delegation, '__init__',
            return_value=None,
        )
        mock_Method = mocker.patch.object(
            elements, 'Method',
            return_value=mocker.Mock(ident=None),
        )

        result = elements.mount(delegation)

        assert result == delegation
        assert not hasattr(result, '_micropath_methods')
        mock_init.assert_not_called()
        mock_Method.assert_not_called()
